import torch
from torch import nn
from torch import Tensor
import torch.nn.functional as F
from torch.autograd import Variable


def focal_loss_with_smoothing(logits, label, weight=None, gamma=0.0, label_smoothing=0.0):
    """ 
    Adapted from https://github.com/Kageshimasu/focal-loss-with-smoothing 
    and https://github.com/clcarwin/focal_loss_pytorch
    """
    log_probabilities = F.log_softmax(logits, dim=-1)
    label = label.view(-1,1)
    log_probability = log_probabilities.gather(1,label).squeeze()
    n_classes = logits.size(1)
    uniform_probability = label_smoothing / n_classes
    label_distribution = torch.full_like(logits, uniform_probability)
    label_distribution.scatter_(1, label, 1.0 - label_smoothing + uniform_probability)

    probability = Variable(log_probability.data.exp())
    difficulty_level = (1-probability)** gamma
    loss = -difficulty_level * torch.sum(log_probabilities * label_distribution, dim=1)

    # Weights
    if weight is not None:
        weight = weight.to(logits.device)
        loss *= torch.gather(weight, -1, label.squeeze())/weight.mean()

    return loss.mean()


class HierarchicalSoftmaxLoss(nn.Module):
    """
    A module which sums the loss for each level of a hiearchical tree.
    """
    def __init__(
        self,
        root,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.root = root

        # Set the indexes of the tree if necessary
        self.root.set_indexes_if_unset()

        assert len(self.root.node_list) > 0

    def forward(self, batch_predictions: Tensor, targets: Tensor) -> Tensor:
        target_nodes = (self.root.node_list[target] for target in targets)

        total_loss = 0.0
        device = targets.device

        for prediction, target_node in zip(batch_predictions, target_nodes):
            node = target_node
            while node.parent:
                # if this is the sole child, then skip it
                if len(node.parent.children) == 1:
                    node = node.parent
                    continue
                
                node.index_in_parent_tensor = node.index_in_parent_tensor.to(device) # can this be done elsewhere?
                logits = torch.unsqueeze(prediction[node.parent.softmax_start_index:node.parent.softmax_end_index], dim=0)
                label = node.index_in_parent_tensor                
                weight = node.parent.weight
                label_smoothing = node.parent.label_smoothing
                gamma = node.parent.gamma
                if gamma is not None and gamma > 0.0:
                    loss = focal_loss_with_smoothing(
                        logits,
                        label,
                        weight=weight,
                        label_smoothing=label_smoothing,
                        gamma=gamma,
                    )
                else:
                    loss = F.cross_entropy(
                        logits,
                        label,
                        weight=weight,
                        label_smoothing=label_smoothing,
                    )

                total_loss += node.parent.alpha * loss
                node = node.parent
                
        batch_size = len(targets)
        total_loss /= batch_size
        return total_loss




# class HierarchicalSoftmaxLoss(nn.Module):
#     """
#     A module which sums the loss for each level of a hiearchical tree.
#     """
#     def __init__(
#         self,
#         root,
#         **kwargs
#     ):
#         super().__init__(**kwargs)
#         self.root = root

#         # Set the indexes of the tree if necessary
#         self.root.set_indexes_if_unset()

#         assert len(self.root.node_list) > 0

#     def forward(self, batch_predictions: Tensor, targets: Tensor) -> Tensor:
#         target_nodes = (self.root.node_list[target] for target in targets)

#         loss = 0.0
#         device = targets.device

#         for prediction, target_node in zip(batch_predictions, target_nodes):
#             node = target_node
#             while node.parent:
#                 node.index_in_parent_tensor = node.index_in_parent_tensor.to(device) # can this be done elsewhere?
#                 print(node.index_in_parent_tensor)
#                 loss += node.parent.alpha * F.cross_entropy(
#                     torch.unsqueeze(prediction[node.parent.softmax_start_index:node.parent.softmax_end_index], dim=0),
#                     node.index_in_parent_tensor,
#                     weight=node.parent.weight,
#                     label_smoothing=node.parent.label_smoothing,
#                 )
#                 node = node.parent
                
#         batch_size = len(targets)
#         loss /= batch_size
#         return loss


# /// script
# [tool.marimo.runtime]
# auto_instantiate = false
# ///

import marimo

__generated_with = "0.13.7"
app = marimo.App(width="full")


@app.cell(hide_code=True)
def _():
    import marimo as mo
    return (mo,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    # CIFAR Demonstration

    This notebook demonstrates how to use the `hierarchicalsoftmax` module to train a neural network on the [CIFAR](https://www.cs.toronto.edu/~kriz/cifar.html) dataset.
    """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""First, choose the hyperparameters.""")
    return


@app.cell
def _(mo):
    cifar_radio = mo.ui.radio(options=["10","100"], value=mo.cli_args().get("cifar") or "100", label="CIFAR Dataset")
    batch_size_input = mo.ui.number(value=mo.cli_args().get("batch") or 32, label="Batch Size")
    epochs_input = mo.ui.number(value=mo.cli_args().get("batch") or 30, label="Epochs")
    mo.vstack([cifar_radio, epochs_input, batch_size_input])
    return batch_size_input, cifar_radio, epochs_input


@app.cell
def _(batch_size_input, cifar_radio, epochs_input):
    from torchvision import datasets, transforms
    from torch.utils.data import DataLoader

    assert cifar_radio.value in ["10","100"]
    batch_size = batch_size_input.value
    epochs = epochs_input.value
    cifar_dataset = datasets.CIFAR10 if cifar_radio.value == "10" else datasets.CIFAR100

    # Use the same data augmentation strategies as in https://arxiv.org/pdf/1605.07146v4
    transform = transforms.Compose([
        transforms.RandomCrop(32, padding=4, padding_mode="reflect"),
        transforms.RandomHorizontalFlip(),
        transforms.ToTensor(),
    ])

    train_data = cifar_dataset(root=".", train=True, download=True, transform=transform)
    test_data = cifar_dataset(root=".", train=False, download=True, transform=transforms.ToTensor())

    train_loader = DataLoader(train_data, batch_size=batch_size, shuffle=True)
    test_loader = DataLoader(test_data, batch_size=batch_size, shuffle=False)
    return (
        DataLoader,
        batch_size,
        epochs,
        test_data,
        test_loader,
        train_data,
        train_loader,
    )


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""### Plot the first 10 images""")
    return


@app.cell
def _(train_data):
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots

    num_images = 10

    # Create a row of subplots
    cifar_fig = make_subplots(
        rows=1, cols=num_images, 
        subplot_titles=[train_data.classes[train_data[i][1]] for i in range(num_images)], 
        horizontal_spacing=0,
    )

    for i in range(num_images):
        img, label = train_data[i]
        img = img.permute(1, 2, 0).numpy()  # (C, H, W) -> (H, W, C) and convert to numpy

        cifar_fig.add_trace(
            go.Image(z=(img * 255).astype('uint8')),
            row=1, col=i+1
        )

    # Update layout: remove axes and tighten spacing
    thumbnail_size = 105
    cifar_fig.update_layout(
        height=thumbnail_size,  # adjust height as needed
        width=thumbnail_size * num_images,  # 150px per image
        showlegend=False,
        margin=dict(l=0, r=0, t=30, b=0)
    )

    # Hide axes
    for i in range(1, num_images + 1):
        cifar_fig.update_xaxes(visible=False, row=1, col=i)
        cifar_fig.update_yaxes(visible=False, row=1, col=i)

    cifar_fig
    return (go,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    ## Non-hierarchical model

    First we create a basic non-hierarchical model as a baseline
    """
    )
    return


@app.cell
def _(train_data):
    import torch
    from torch import nn
    import torch.nn.functional as F
    import lightning as L
    from torchmetrics import Accuracy

    class BasicBlock(nn.Module):
        def __init__(self, in_planes, out_planes, stride, dropout_rate=0.3):
            super().__init__()
            self.bn1 = nn.BatchNorm2d(in_planes)
            self.conv1 = nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False)
            self.dropout = nn.Dropout(p=dropout_rate) if dropout_rate > 0 else nn.Identity()
            self.bn2 = nn.BatchNorm2d(out_planes)
            self.conv2 = nn.Conv2d(out_planes, out_planes, kernel_size=3, stride=1, padding=1, bias=False)

            self.shortcut = nn.Sequential()
            if stride != 1 or in_planes != out_planes:
                self.shortcut = nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)

        def forward(self, x):
            out = self.conv1(F.relu(self.bn1(x)))
            out = self.dropout(out)
            out = self.conv2(F.relu(self.bn2(out)))
            out += self.shortcut(x)
            return out


    class WideResNetBody(nn.Module):
        def __init__(self, depth:int=28, width_factor:int=10, dropout_rate:float=0.3):
            super().__init__()
            assert (depth - 4) % 6 == 0, "Depth should be 6n+4"
            n = (depth - 4) // 6

            k = width_factor
            self.in_planes = 16

            self.conv1 = nn.Conv2d(3, 16, kernel_size=3, stride=1, padding=1, bias=False)

            self.layer1 = self._make_layer(16*k, n, stride=1, dropout_rate=dropout_rate)
            self.layer2 = self._make_layer(32*k, n, stride=2, dropout_rate=dropout_rate)
            self.layer3 = self._make_layer(64*k, n, stride=2, dropout_rate=dropout_rate)

            self.bn = nn.BatchNorm2d(64*k)

        def _make_layer(self, out_planes, blocks, stride, dropout_rate):
            strides = [stride] + [1]*(blocks-1)
            layers = []
            for s in strides:
                layers.append(BasicBlock(self.in_planes, out_planes, s, dropout_rate))
                self.in_planes = out_planes
            return nn.Sequential(*layers)

        def forward(self, x):
            out = self.conv1(x)
            out = self.layer1(out)
            out = self.layer2(out)
            out = self.layer3(out)
            out = F.relu(self.bn(out))
            out = F.avg_pool2d(out, 8)
            out = out.view(out.size(0), -1)
            return out


    class BasicImageClassifier(L.LightningModule):
        def __init__(self, depth:int=28, width_factor:int=10, dropout_rate:float=0.3):
            super().__init__()
            self.model = nn.Sequential(
                WideResNetBody(depth=depth, width_factor=width_factor, dropout_rate=dropout_rate),
                nn.LazyLinear(out_features=len(train_data.classes))
            )
            self.loss_fn = nn.CrossEntropyLoss()
            self.metrics = [
                Accuracy(task="multiclass", num_classes=len(train_data.classes))
            ]

        def forward(self, x):
            return self.model(x)

        def training_step(self, batch, batch_idx):
            x, y = batch
            logits = self(x)
            loss = self.loss_fn(logits, y)
            self.log('train_loss', loss, prog_bar=True)
            return loss

        def validation_step(self, batch, batch_idx):
            x, y = batch
            logits = self(x)
            loss = self.loss_fn(logits, y)
            self.log('val_loss', loss, prog_bar=True)
            for metric in self.metrics:
                metric = metric.to(logits.device)
                result = metric(logits, y)
                if isinstance(result, dict):
                    for name, value in result.items():
                        self.log(f"val_{name}", value, on_step=False, on_epoch=True, prog_bar=True)
                else:
                    self.log(f"val_{metric.__class__.__name__}", result, on_step=False, on_epoch=True, prog_bar=True)

            return loss

        def configure_optimizers(self):
            return torch.optim.Adam(self.parameters(), lr=1e-3)

    basic_model = BasicImageClassifier(depth=16, width_factor=8, dropout_rate=0.3)
    basic_model
    return BasicImageClassifier, L, WideResNetBody, basic_model, nn, torch


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## Train the basic model""")
    return


@app.cell
def _(L, basic_model, epochs, test_loader, train_loader):
    from lightning.pytorch.loggers import CSVLogger
    from lightning.pytorch.callbacks import TQDMProgressBar

    basic_logger = CSVLogger(save_dir="lightning_logs", name="basic_model")
    basic_trainer = L.Trainer(max_epochs=epochs, accelerator="auto", enable_checkpointing=False, logger=basic_logger, callbacks=[TQDMProgressBar(leave=True, refresh_rate=20)])
    basic_trainer.fit(basic_model, train_dataloaders=train_loader, val_dataloaders=test_loader)
    return CSVLogger, TQDMProgressBar, basic_logger


@app.cell
def _(basic_logger, go):
    import pandas as pd
    from pathlib import Path
    import plotly.io as pio

    pio.templates.default = "plotly_white"

    basic_metrics_df = pd.read_csv(Path(basic_logger.log_dir) / "metrics.csv")
    basic_metrics_df = basic_metrics_df.dropna(subset=["val_MulticlassAccuracy"])
    basic_fig = go.Figure()
    basic_fig.add_trace(go.Scatter(x=basic_metrics_df["epoch"], y=basic_metrics_df["val_MulticlassAccuracy"], mode='lines', name='class'))
    basic_fig.update_layout(
        xaxis_title="Epochs",
        yaxis_title="Accuracy",
    )
    basic_fig.write_html(str(Path(basic_logger.log_dir)/"accuracy.html"))
    basic_fig.show()

    return Path, pd


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        """
    ## Hierarchical model

    Let's now create a hierarchical model.
    First we need to create a tree structure for the CIFAR dataset.
    """
    )
    return


@app.cell
def _(mo, train_data):
    from hierarchicalsoftmax import (
        SoftmaxNode,
        HierarchicalSoftmaxLazyLinear,
        HierarchicalSoftmaxLoss,
    )

    if len(train_data.classes) == 10:
        # CIFAR-10
        superclasses = {
            "animals": ["bird", "cat", "deer", "dog", "frog", "horse"],
            "vehicles": ["airplane", "automobile", "ship", "truck"],
        }
    else:
        # CIFAR-100
        superclasses = {
            "aquatic mammals": ["beaver", "dolphin", "otter", "seal", "whale"],
            "fish": ["aquarium_fish", "flatfish", "ray", "shark", "trout"],
            "flowers": ["orchid", "poppy", "rose", "sunflower", "tulip"],
            "food containers": ["bottle", "bowl", "can", "cup", "plate"],
            "fruit and vegetables": ["apple", "mushroom", "orange", "pear", "sweet_pepper"],
            "household electrical devices": ["clock", "keyboard", "lamp", "telephone", "television"],
            "household furniture": ["bed", "chair", "couch", "table", "wardrobe"],
            "insects": ["bee", "beetle", "butterfly", "caterpillar", "cockroach"],
            "large carnivores": ["bear", "leopard", "lion", "tiger", "wolf"],
            "large man-made outdoor things": ["bridge", "castle", "house", "road", "skyscraper"],
            "large natural outdoor scenes": ["cloud", "forest", "mountain", "plain", "sea"],
            "large omnivores and herbivores": ["camel", "cattle", "chimpanzee", "elephant", "kangaroo"],
            "medium-sized mammals": ["fox", "porcupine", "possum", "raccoon", "skunk"],
            "non-insect invertebrates": ["crab", "lobster", "snail", "spider", "worm"],
            "people": ["baby", "boy", "girl", "man", "woman"],
            "reptiles": ["crocodile", "dinosaur", "lizard", "snake", "turtle"],
            "small mammals": ["hamster", "mouse", "rabbit", "shrew", "squirrel"],
            "trees": ["maple_tree", "oak_tree", "palm_tree", "pine_tree", "willow_tree"],
            "vehicles 1": ["bicycle", "bus", "motorcycle", "pickup_truck", "train"],
            "vehicles 2": ["lawn_mower", "rocket", "streetcar", "tank", "tractor"],
        }


    root = SoftmaxNode("root")
    for superclass, classes in superclasses.items():
        superclass_node = SoftmaxNode(superclass, parent=root)
        for class_name in classes:
            SoftmaxNode(class_name, parent=superclass_node)

    # Now that the tree is built, we can set the indexes
    # This makes the tree read-only
    root.set_indexes()
    name_to_node_id = {node.name: root.node_to_id[node] for node in root.leaves}
    index_to_node_id = {
        i: name_to_node_id[name] for i, name in enumerate(train_data.classes)
    }

    # Render the hierarchy
    mo.Html(root.svg())
    return (
        HierarchicalSoftmaxLazyLinear,
        HierarchicalSoftmaxLoss,
        SoftmaxNode,
        index_to_node_id,
        root,
    )


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""### Create DataLoaders with hierarchical labels""")
    return


@app.cell
def _(DataLoader, batch_size, index_to_node_id, test_data, torch, train_data):
    class HierarchicalDataset(torch.utils.data.Dataset):
        def __init__(self, dataset, index_to_node_id):
            self.dataset = dataset
            self.index_to_node_id = index_to_node_id

        def __getitem__(self, idx):
            image, label = self.dataset[idx]
            return image, self.index_to_node_id[label]

        def __len__(self):
            return len(self.dataset)

    hierarchical_train_loader = DataLoader(HierarchicalDataset(train_data, index_to_node_id), batch_size=batch_size, shuffle=True)
    hierarchical_test_loader = DataLoader(HierarchicalDataset(test_data, index_to_node_id), batch_size=batch_size, shuffle=False)
    return hierarchical_test_loader, hierarchical_train_loader


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""### Create the Hierarchical Image Classifier model""")
    return


@app.cell
def _(
    BasicImageClassifier,
    HierarchicalSoftmaxLazyLinear,
    HierarchicalSoftmaxLoss,
    SoftmaxNode,
    WideResNetBody,
    nn,
    root,
):
    from hierarchicalsoftmax.metrics import RankAccuracyTorchMetric, LeafAccuracyTorchMetric

    class HierarchicalImageClassifier(BasicImageClassifier):
        # Just overriding the init - keep the rest of the code
        def __init__(self, root: SoftmaxNode, depth:int=28, width_factor:int=10, dropout_rate:float=0.3):
            super().__init__()
            self.model = nn.Sequential(
                WideResNetBody(depth=depth, width_factor=width_factor, dropout_rate=dropout_rate),
                HierarchicalSoftmaxLazyLinear(root=root)
            )
            self.loss_fn = HierarchicalSoftmaxLoss(root)
            self.metrics = [
                RankAccuracyTorchMetric(
                    root,
                    {1: "superclass_accuracy"},
                ),
                LeafAccuracyTorchMetric(root, name="class_accuracy"),
            ]
            self.root = root

    hierarchical_model = HierarchicalImageClassifier(root, depth=16, width_factor=8, dropout_rate=0.3)        
    hierarchical_model
    return (hierarchical_model,)


@app.cell
def _(
    CSVLogger,
    L,
    TQDMProgressBar,
    epochs,
    hierarchical_model,
    hierarchical_test_loader,
    hierarchical_train_loader,
):
    hierarchical_logger = CSVLogger(save_dir="lightning_logs", name="hierarchical_model")
    hierarchical_trainer = L.Trainer(max_epochs=epochs, accelerator="auto", enable_checkpointing=False, logger=hierarchical_logger, callbacks=[TQDMProgressBar(leave=True, refresh_rate=20)])
    hierarchical_trainer.fit(hierarchical_model, train_dataloaders=hierarchical_train_loader, val_dataloaders=hierarchical_test_loader)
    return (hierarchical_logger,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""### Plot the validation results at both the superclass and the class levels""")
    return


@app.cell
def _(Path, go, hierarchical_logger, pd):
    hierarchical_df = pd.read_csv(Path(hierarchical_logger.log_dir) / "metrics.csv")
    hierarchical_df = hierarchical_df.dropna(subset=["val_LeafAccuracyTorchMetric"])
    hierarchical_fig = go.Figure()
    hierarchical_fig.add_trace(go.Scatter(x=hierarchical_df["epoch"], y=hierarchical_df["val_superclass_accuracy"], mode='lines', name='superclass'))
    hierarchical_fig.add_trace(go.Scatter(x=hierarchical_df["epoch"], y=hierarchical_df["val_LeafAccuracyTorchMetric"], mode='lines', name='class'))
    hierarchical_fig.update_layout(
        xaxis_title="Epochs",
        yaxis_title="Accuracy",
    )
    hierarchical_fig.write_html(str(Path(hierarchical_logger.log_dir)/"accuracy.html"))
    hierarchical_fig
    return


@app.cell(hide_code=True)
def _():
    return


if __name__ == "__main__":
    app.run()



from torch import Tensor, Size
from functools import cached_property
import torch.nn.functional as F
from torch.nn.parameter import Parameter


class LazyLinearTensor(Tensor):
    """
    A tensor that is designed to be used with HierarchicalSoftmaxLazyLinear layers.
    """
    @staticmethod
    def __new__(cls, x, weight:Parameter, bias:Parameter, *args, **kwargs):
        return super().__new__(cls, x, *args, **kwargs)

    def __init__(self, x:Tensor, weight:Parameter, bias:Parameter, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.input = x
        self.weight = weight
        self.bias = bias

    @cached_property
    def result(self):
        return F.linear(self.input, self.weight, self.bias)

    def __add__(self, other):
        return self.result + other
    
    def __sub__(self, other):
        return self.result - other
    
    def __mul__(self, other):
        return self.result * other
    
    def __truediv__(self, other):
        return self.result / other
    
    def __matmul__(self, other):
        return self.result @ other
    
    def __radd__(self, other):
        return other + self.result
    
    def __rsub__(self, other):
        return other - self.result
    
    def __rmul__(self, other):
        return other * self.result
    
    def __rtruediv__(self, other):
        return other / self.result
    
    def __rmatmul__(self, other):
        return other @ self.result

    def __getitem__(self, index):
        assert isinstance(index, int) or isinstance(index, slice) or isinstance(index, tuple)
        if not isinstance(index, tuple) or isinstance(index, slice):
            index = (index,)

        my_shape = self.shape
        if len(index) < len(my_shape):
            return LazyLinearTensor(self.input[index], weight=self.weight, bias=self.bias)
        if len(index) > len(my_shape):
            raise IndexError(f"Cannot get index '{index}' for LazyLinearTensor of shape {len(my_shape)}")

        input = self.input[index[:-1]]
        weight = self.weight[index[-1]]
        bias = self.bias[index[-1]]      
        return F.linear(input, weight, bias)
    
    @property
    def shape(self) -> Size:
        return Size( self.input.shape[:-1] + (self.weight.shape[0],) )

    def __str__(self) -> str:
        return f"LazyLinearTensor (shape={tuple(self.shape)})"
    
    def __repr__(self) -> str:
        return str(self)
    
    def __len__(self) -> int:
        return self.shape[0]
    
    def __iter__(self):
        for i in range(len(self)):
            yield self[i]
    
    def float(self):
        x = super().float()
        x.input = self.input.float()
        x.weight = self.weight.float()
        x.bias = self.bias.float()
        return x
    
    def half(self):
        x = super().half()
        x.input = self.input.half()
        x.weight = self.weight.half()
        x.bias = self.bias.half()
        return x    
    
    def to(self, *args, **kwargs):
        x = super().to(*args, **kwargs)
        x.input = self.input.to(*args, **kwargs)
        x.weight = self.weight.to(*args, **kwargs)
        x.bias = self.bias.to(*args, **kwargs)
        return x



from anytree.exporter import DotExporter
from anytree import PreOrderIter

class ThresholdDotExporter(DotExporter):
    def __init__(
        self, 
        node, 
        probabilities, 
        greedy_nodes, 
        graph="digraph", 
        name="tree", 
        options=None,
        indent=4, 
        nodenamefunc=None, 
        nodeattrfunc=None,
        edgeattrfunc=None, 
        edgetypefunc=None,
        prediction_color="red",
        non_prediction_color="gray",
        horizontal:bool=True,
        threshold:float=0.005,
    ):
        options = options or []
        if horizontal:
            options.append('rankdir="LR";')
        
        super().__init__(
            node, 
            graph=graph, 
            name=name, 
            options=options, 
            indent=indent,
            nodenamefunc=nodenamefunc, 
            nodeattrfunc=nodeattrfunc,
            edgeattrfunc=edgeattrfunc, 
            edgetypefunc=edgetypefunc
        )
        self.greedy_nodes = greedy_nodes
        self.probabilities = probabilities
        self.prediction_color = prediction_color
        self.non_prediction_color = non_prediction_color
        self.threshold = threshold
        self.excluded_nodes = set()

    def _default_nodeattrfunc(self, node):
        return f"color={self.prediction_color}" if node in self.greedy_nodes else ""

    def _default_edgeattrfunc(
        self,
        parent, 
        child,
    ):
        color = self.prediction_color if child in self.greedy_nodes else self.non_prediction_color
        label = f"{self.probabilities[child.index_in_softmax_layer]:.2f}" if child.index_in_softmax_layer is not None else "x"
        return f"label={label},color={color}"

    def exclude_node(self, node) -> bool:
        if node in self.excluded_nodes:
            return True
        
        if node.index_in_softmax_layer is None:
            exclude_node = node.parent in self.excluded_nodes
        else:
            include_node = node.is_root or node in self.greedy_nodes or self.probabilities[node.index_in_softmax_layer] >= self.threshold
            exclude_node = not include_node
            
        if exclude_node:
            self.excluded_nodes.add(node)
        return exclude_node

    def _DotExporter__iter_nodes(self, indent, nodenamefunc, nodeattrfunc, *args, **kwargs):
        for node in PreOrderIter(self.node, maxlevel=self.maxlevel):
            if self.exclude_node(node):
                continue
            nodename = nodenamefunc(node)
            nodeattr = nodeattrfunc(node)
            nodeattr = " [%s]" % nodeattr if nodeattr is not None else ""
            yield '%s"%s"%s;' % (indent, DotExporter.esc(nodename), nodeattr)

    def _DotExporter__iter_edges(self, indent, nodenamefunc, edgeattrfunc, edgetypefunc, *args, **kwargs):
        maxlevel = self.maxlevel - 1 if self.maxlevel else None
        for node in PreOrderIter(self.node, maxlevel=maxlevel):
            nodename = nodenamefunc(node)
            for child in node.children:
                if self.exclude_node(child):
                    continue

                childname = nodenamefunc(child)
                edgeattr = edgeattrfunc(node, child)
                edgetype = edgetypefunc(node, child)
                edgeattr = " [%s]" % edgeattr if edgeattr is not None else ""
                yield '%s"%s" %s "%s"%s;' % (indent, DotExporter.esc(nodename), edgetype,
                                             DotExporter.esc(childname), edgeattr)



from .nodes import SoftmaxNode
from .loss import HierarchicalSoftmaxLoss
from .metrics import greedy_accuracy, greedy_f1_score
from .inference import greedy_predictions
from .layers import HierarchicalSoftmaxLinear, HierarchicalSoftmaxLazyLinear
from .treedict import TreeDict


import copy
from pathlib import Path
from attrs import define, field
from collections import UserDict
import pickle
from collections import Counter
from rich.progress import track
import typer

from .nodes import SoftmaxNode


@define
class NodeDetail:
    """
    Stores metadata for a key in the TreeDict.

    Attributes:
        partition (int): The partition ID this key belongs to.
        node (SoftmaxNode): The node in the classification tree associated with the key.
        node_id (int): The index of the node in the tree (used during pickling).
    """
    partition:int
    node:SoftmaxNode = field(default=None, eq=False)
    node_id:int = None

    def __getstate__(self):
        return (self.partition, self.node_id)

    def __setstate__(self, state):
        self.partition, self.node_id = state
        self.node = None


class AlreadyExists(Exception):
    pass


class TreeDict(UserDict):
    def __init__(self, classification_tree:SoftmaxNode|None=None):
        """
        Initialize a TreeDict.

        Args:
            classification_tree (SoftmaxNode, optional): The root of the classification tree.
                If not provided, a new root node named "root" will be created.
        """
        super().__init__()
        self.classification_tree = classification_tree or SoftmaxNode("root")

    def add(self, key:str, node:SoftmaxNode, partition:int) -> NodeDetail:
        """
        Associate a key with a node and a partition.

        Args:
            key (str): The unique identifier for the item.
            node (SoftmaxNode): The node in the classification tree to associate with the key.
            partition (int): The partition index for the key.

        Raises:
            AlreadyExists: If the key already exists with a different node.

        Returns:
            NodeDetail: The metadata object for the added key.
        """
        assert node.root == self.classification_tree
        if key in self:
            old_node = self.node(key)
            if not node == old_node:
                raise AlreadyExists(f"Accession {key} already exists in TreeDict at node {self.node(key)}. Cannot change to {node}")

        detail = NodeDetail(
            partition=partition,
            node=node,
        )
        self[key] = detail
        return detail

    def set_indexes(self):
        """
        Ensure the tree has assigned node indexes, and record the node_id for each key.
        """
        self.classification_tree.set_indexes_if_unset()
        for detail in self.values():
            if detail.node:
                detail.node_id = self.classification_tree.node_to_id[detail.node]

    def save(self, path:Path):
        """
        Save the TreeDict to a pickle file.

        Args:
            path (Path): The file path to save the TreeDict.
        """
        path = Path(path)
        path.parent.mkdir(exist_ok=True, parents=True)

        self.set_indexes()
        with open(path, 'wb') as handle:
            pickle.dump(self, handle, protocol=pickle.HIGHEST_PROTOCOL)

    @classmethod
    def load(self, path:Path):
        """
        Load a TreeDict from a pickle file.

        Args:
            path (Path): The path to the serialized TreeDict.

        Returns:
            TreeDict: The loaded TreeDict instance.
        """
        with open(path, 'rb') as handle:
            return pickle.load(handle)

    def node(self, key:str):
        """
        Retrieve the node associated with a key.

        Args:
            key (str): The key for which to retrieve the node.

        Returns:
            SoftmaxNode: The node corresponding to the key.
        """
        detail = self[key]
        if detail.node is not None:
            return detail.node
        return self.classification_tree.node_list[detail.node_id]
    
    def keys_in_partition(self, partition:int):
        """
        Yield all keys that belong to a given partition.

        Args:
            partition (int): The partition to filter by.

        Yields:
            str: Keys in the specified partition.
        """
        for key, detail in self.items():
            if detail.partition == partition:
                yield key

    def keys(self, partition:int|None = None):
        """
        Return keys in the TreeDict, optionally filtering by partition.

        Args:
            partition (int | None): The partition to filter keys by. If None, return all keys.

        Returns:
            Iterator[str]: An iterator over the keys.
        """
        return super().keys() if partition is None else self.keys_in_partition(partition)
    
    def truncate(self, max_depth:int) -> "TreeDict":
        """
        Truncate the classification tree to a specified maximum depth and return a new TreeDict.

        Keys deeper than the depth limit will be reassigned to the ancestor node at that depth.

        Args:
            max_depth (int): The maximum number of ancestor levels to keep.

        Returns:
            TreeDict: A new truncated TreeDict.
        """
        self.classification_tree.set_indexes_if_unset()
        classification_tree = copy.deepcopy(self.classification_tree)
        new_treedict = TreeDict(classification_tree)
        for key in track(self.keys()):
            original_node = self.node(key)
            node_id = self.classification_tree.node_to_id[original_node]
            node = classification_tree.node_list[node_id]

            ancestors = node.ancestors
            if len(ancestors) >= max_depth:
                node = ancestors[max_depth-1]
            new_treedict.add(key, node, self[key].partition)

        # Remove any nodes that beyond the max depth
        for node in new_treedict.classification_tree.pre_order_iter():
            node.readonly = False
            node.softmax_start_index = None
            if len(node.ancestors) >= max_depth:
                node.parent = None

        new_treedict.set_indexes()

        return new_treedict
                
    def add_counts(self):
        """
        Count the number of keys assigned to each node, and store the count in each node.
        """
        for node in self.classification_tree.post_order_iter():
            node.count = 0

        for key in self.keys():
            node = self.node(key)
            node.count += 1

    def add_partition_counts(self):
        """
        Count the number of keys in each partition per node and store it in the node.
        """
        for node in self.classification_tree.post_order_iter():
            node.partition_counts = Counter()

        for key, detail in self.items():
            node = self.node(key)
            partition = detail.partition
            node.partition_counts[partition] += 1

    def render(self, count:bool=False, partition_counts:bool=False, **kwargs):
        """
        Render the tree as text, optionally showing key counts or partition counts.

        Args:
            count (bool): If True, show the number of keys at each node.
            partition_counts (bool): If True, show partition-wise key counts at each node.
            **kwargs: Additional arguments passed to the underlying tree render method.

        Returns:
            anytree.RenderTree or str: The rendered tree.
        """
        if partition_counts:
            self.add_partition_counts()
            for node in self.classification_tree.post_order_iter():
                partition_counts_str = "; ".join([f"{k}->{node.partition_counts[k]}" for k in sorted(node.partition_counts.keys())])
                node.render_str = f"{node.name} {partition_counts_str}"
            kwargs['attr'] = "render_str"
        elif count:
            self.add_counts()
            for node in self.classification_tree.post_order_iter():
                node.render_str = f"{node.name} ({node.count})" if getattr(node, "count", 0) else node.name

            kwargs['attr'] = "render_str"

        return self.classification_tree.render(**kwargs)
    
    def sunburst(self, **kwargs) -> "go.Figure":
        """
        Generate a Plotly sunburst plot based on the TreeDict.

        Node values are based on the number of keys mapped to each node.

        Args:
            **kwargs: Additional keyword arguments passed to Plotly layout.

        Returns:
            plotly.graph_objects.Figure: A sunburst plot.
        """
        import plotly.graph_objects as go

        self.add_counts()
        labels = []
        parents = []
        values = []

        for node in self.classification_tree.pre_order_iter():
            labels.append(node.name)
            parents.append(node.parent.name if node.parent else "")
            values.append(node.count)

        fig = go.Figure(go.Sunburst(
            labels=labels,
            parents=parents,
            values=values,
            branchvalues="remainder",
        ))
        
        fig.update_layout(margin=dict(t=10, l=10, r=10, b=10), **kwargs)
        return fig

    def keys_to_file(self, file:Path) -> None:
        """
        Write all keys to a text file, one per line.

        Args:
            file (Path): Path to the output text file.
        """
        with open(file, "w") as f:
            for key in self.keys():
                print(key, file=f)

    def csv(self, file:Path) -> None:
        """
        Write all keys, node names and partitions to a CSV file.

        Args:
            file (Path): Path to the output text file.
        """
        with open(file, "w") as f:
            print("key,node,partition", file=f)
            for key in self.keys():
                detail = self[key]
                node = self.node(key)
                print(f"{key},{node.name.strip()},{detail.partition}", file=f)

    def pickle_tree(self, output:Path):
        """
        Save only the classification tree (not the key-to-node mapping) to a pickle file.

        Args:
            output (Path): Path to the output file.
        """
        with open(output, 'wb') as pickle_file:
            pickle.dump(self.classification_tree, pickle_file)


app = typer.Typer()

@app.command()
def keys(
    treedict:Path = typer.Argument(...,help="The path to the TreeDict."), 
    partition:int|None = typer.Option(None,help="The index of the partition to list."), 
):
    """ 
    Prints a list of keys in a TreeDict. 
    
    If a partition is given, then only the keys for that partition are given.
    """
    treedict = TreeDict.load(treedict)
    for key in treedict.keys(partition=partition):
        print(key)
    

@app.command()
def csv(
    treedict:Path = typer.Argument(...,help="The path to the TreeDict."), 
    csv:Path = typer.Argument(...,help="The path to the output CSV file."),
):
    """ 
    Writes a CSV file with the key, node name and partition. 
    """
    treedict = TreeDict.load(treedict)
    treedict.csv(csv)
    

@app.command()
def render(
    treedict:Path = typer.Argument(...,help="The path to the TreeDict."), 
    output:Path|None = typer.Option(None, help="The path to save the rendered tree."),
    print_tree:bool = typer.Option(True, help="Whether or not to print the tree to the screen."),
    count:bool = typer.Option(False, help="Whether or not to print the count of keys at each node."),
    partition_counts:bool = typer.Option(False, help="Whether or not to print the count of each partition at each node."),
):
    """
    Render the tree as text, optionally showing key counts or partition counts.
    """
    treedict = TreeDict.load(treedict)
    treedict.render(filepath=output, print=print_tree, count=count, partition_counts=partition_counts)


@app.command()
def count(
    treedict:Path = typer.Argument(...,help="The path to the TreeDict."), 
):
    """
    Prints the number of keys in the TreeDict.
    """
    treedict = TreeDict.load(treedict)
    print(len(treedict))


@app.command()
def sunburst(
    treedict:Path = typer.Argument(...,help="The path to the TreeDict."), 
    show:bool = typer.Option(False, help="Whether or not to show the plot."),
    output:Path = typer.Option(None, help="The path to save the rendered tree."),
    width:int = typer.Option(1000, help="The width of the plot."),
    height:int = typer.Option(0, help="The height of the plot. If 0 then it will be calculated based on the width."),
):
    """
    Renders the TreeDict as a sunburst plot.
    """
    treedict = TreeDict.load(treedict)
    height = height or width

    fig = treedict.sunburst(width=width, height=height)
    if show:
        fig.show()
    
    if output:
        output = Path(output)
        output.parent.mkdir(exist_ok=True, parents=True)

        # if kaleido is installed, turn off mathjax
        # https://github.com/plotly/plotly.py/issues/3469
        try:
            import plotly.io as pio
            pio.kaleido.scope.mathjax = None
        except Exception as e:
            pass

        output_func = fig.write_html if output.suffix.lower() == ".html" else fig.write_image
        output_func(output)


@app.command()
def truncate(
    treedict:Path = typer.Argument(...,help="The path to the TreeDict."), 
    max_depth:int = typer.Argument(...,help="The maximum depth to truncate the tree."),
    output:Path = typer.Argument(...,help="The path to the output file."),
):
    """
    Truncates the tree to a maximum depth.
    """
    treedict = TreeDict.load(treedict)
    new_tree = treedict.truncate(max_depth)
    new_tree.save(output)


@app.command()
def layer_size(
    treedict:Path = typer.Argument(...,help="The path to the TreeDict."),         
):
    """
    Prints the size of the neural network layer to predict the classification tree.
    """    
    treedict = TreeDict.load(treedict)
    print(treedict.classification_tree.layer_size)


@app.command()
def pickle_tree(
    treedict:Path = typer.Argument(...,help="The path to the TreeDict."),    
    output:Path = typer.Argument(...,help="The path to the output pickle file."),     
):
    """
    Pickles the classification tree to a file.
    """
    treedict = TreeDict.load(treedict)
    treedict.pickle_tree(output)



from typing import List, Optional
import torch
from pathlib import Path
from anytree import PreOrderIter
from rich.progress import track

from . import nodes
from .dotexporter import ThresholdDotExporter


class ShapeError(RuntimeError):
    """
    Raised when the shape of a tensor is different to what is expected.
    """

def node_probabilities(prediction_tensor:torch.Tensor, root:nodes.SoftmaxNode, progress_bar:bool=False) -> torch.Tensor:
    """
    Takes the prediction scores for a number of samples and converts it to a list of probabilities of nodes in the tree.
    """
    probabilities = torch.zeros(size=prediction_tensor.shape, device=prediction_tensor.device)

    if root.softmax_start_index is None:
        raise nodes.IndexNotSetError(f"The index of the root node {root} has not been set. Call `set_indexes` on this object.")

    if prediction_tensor.shape[-1] != root.layer_size:
        raise ShapeError(
            f"The predictions tensor given to {__name__} has final dimensions of {prediction_tensor.shape[-1]}. "
            f"That is not compatible with the root node which expects prediciton tensors to have a final dimension of {root.layer_size}."
        )

    for node in track(PreOrderIter(root)) if progress_bar else PreOrderIter(root):
        if node.is_leaf:
            continue
        elif node == root:
            my_probability = 1.0
        elif node.index_in_softmax_layer != None:
            my_probability = probabilities[:,node.index_in_softmax_layer]
            my_probability = my_probability[:,None]

        if len(node.children) == 1:
            # If this has just one child, then skip it
            continue

        softmax_probabilities = torch.softmax(
            prediction_tensor[:,node.softmax_start_index:node.softmax_end_index], 
            dim=1,
        )

        probabilities[:,node.softmax_start_index:node.softmax_end_index] = softmax_probabilities * my_probability
    
    return probabilities


def leaf_probabilities(prediction_tensor:torch.Tensor, root:nodes.SoftmaxNode) -> torch.Tensor:
    """
    Takes the prediction scores for a number of samples and converts it to a list of probabilities of nodes in the tree.
    """
    probabilities = node_probabilities(prediction_tensor, root=root)
    return torch.index_select(probabilities, 1, root.leaf_indexes.to(probabilities.device))


def greedy_predictions(
        prediction_tensor:torch.Tensor, 
        root:nodes.SoftmaxNode, 
        max_depth:Optional[int]=None, 
        threshold:Optional[float]=None,
        progress_bar:bool=False,
    ) -> List[nodes.SoftmaxNode]:
    """
    Takes the prediction scores for a number of samples and converts it to a list of predictions of nodes in the tree.

    Predictions use the `greedy` method which means that it chooses the greatest prediction score at each level of the tree.

    Args:
        prediction_tensor (torch.Tensor): The output from the softmax layer. 
            Shape (samples, root.layer_size)
            Works with raw scores or probabilities.
        root (SoftmaxNode): The root softmax node. Needs `set_indexes` to have been called.
        prediction_tensor (torch.Tensor): The predictions coming from the softmax layer. Shape (samples, root.layer_size)
        max_depth (int, optional): If set, then it only gives predictions at a maximum of this number of levels from the root.
        threshold (int, optional): If set, then it only gives predictions where the value at the node is greater than this threshold.
            Designed for use with probabilities.

    Returns:
        List[nodes.SoftmaxNode]: A list of nodes predicted for each sample.
    """
    prediction_nodes = []

    if isinstance(prediction_tensor, tuple) and len(prediction_tensor) == 1:
        prediction_tensor = prediction_tensor[0]

    if root.softmax_start_index is None:
        raise nodes.IndexNotSetError(f"The index of the root node {root} has not been set. Call `set_indexes` on this object.")

    if prediction_tensor.shape[-1] != root.layer_size:
        raise ShapeError(
            f"The predictions tensor given to {__name__} has final dimensions of {prediction_tensor.shape[-1]}. "
            f"That is not compatible with the root node which expects prediciton tensors to have a final dimension of {root.layer_size}."
        )

    for predictions in track(prediction_tensor) if progress_bar else prediction_tensor:
        node = root
        depth = 1
        while (node.children):
            if len(node.children) == 1:
                # if this has just one child, then we don't check the prediction
                prediction_child_index = 0
            else:
                # This would be better if we could use torch.argmax but it doesn't work with MPS in the production version of pytorch
                # See https://github.com/pytorch/pytorch/issues/98191
                # https://github.com/pytorch/pytorch/pull/104374
                prediction_child_index = torch.max(predictions[node.softmax_start_index:node.softmax_end_index], dim=0).indices

            # Stop if the prediction is below the threshold
            if threshold and predictions[node.softmax_start_index+prediction_child_index] < threshold:
                break
            
            node = node.children[prediction_child_index]

            # Stop if we have reached the maximum depth
            if max_depth and depth >= max_depth:
                break

            depth += 1

        prediction_nodes.append(node)
    
    return prediction_nodes


def greedy_lineage_probabilities(
        prediction_tensor:torch.Tensor, 
        root:nodes.SoftmaxNode, 
        max_depth:Optional[int]=None, 
        threshold:Optional[float]=None,
        progress_bar:bool=False,
    ) -> List[nodes.SoftmaxNode]:
    """
    Takes the prediction scores for a number of samples and converts it to a list of predictions of nodes in the tree.

    Predictions use the `greedy` method which means that it chooses the greatest prediction score at each level of the tree.

    Args:
        prediction_tensor (torch.Tensor): The output from the softmax layer. 
            Shape (samples, root.layer_size)
            Works with raw scores or probabilities.
        root (SoftmaxNode): The root softmax node. Needs `set_indexes` to have been called.
        prediction_tensor (torch.Tensor): The predictions coming from the softmax layer. Shape (samples, root.layer_size)
        max_depth (int, optional): If set, then it only gives predictions at a maximum of this number of levels from the root.
        threshold (int, optional): If set, then it only gives predictions where the value at the node is greater than this threshold.
            Designed for use with probabilities.

    Returns:
        List[List[Tuple[nodes.SoftmaxNode, float]]]: A list of nodes predicted for each sample with their probabilities.
    """
    prediction_lineages = []

    if isinstance(prediction_tensor, tuple) and len(prediction_tensor) == 1:
        prediction_tensor = prediction_tensor[0]

    if root.softmax_start_index is None:
        raise nodes.IndexNotSetError(f"The index of the root node {root} has not been set. Call `set_indexes` on this object.")

    if prediction_tensor.shape[-1] != root.layer_size:
        raise ShapeError(
            f"The predictions tensor given to {__name__} has final dimensions of {prediction_tensor.shape[-1]}. "
            f"That is not compatible with the root node which expects prediciton tensors to have a final dimension of {root.layer_size}."
        )

    for predictions in track(prediction_tensor) if progress_bar else prediction_tensor:
        node = root
        depth = 1
        probability = 1.0 # Start with the root node having a probability of 1.0
        my_lineage = []
        while (node.children):
            if len(node.children) == 1:
                # if this has just one child, then we don't check the prediction
                prediction_child_index = 0
            else:
                # This would be better if we could use torch.argmax but it doesn't work with MPS in the production version of pytorch
                # See https://github.com/pytorch/pytorch/issues/98191
                # https://github.com/pytorch/pytorch/pull/104374
                child_probabilities = torch.softmax(
                    predictions[node.softmax_start_index:node.softmax_end_index], 
                    dim=0,
                )
                prediction_child_index = torch.max(child_probabilities, dim=0).indices
                probability *= child_probabilities[prediction_child_index].item()

            # Stop if the probability is below the threshold
            if threshold and probability < threshold:
                break
            
            node = node.children[prediction_child_index]

            my_lineage.append(
                (node, probability)
            )

            # Stop if we have reached the maximum depth
            if max_depth and depth >= max_depth:
                break

            depth += 1

        prediction_lineages.append(my_lineage)
    
    return prediction_lineages


def greedy_prediction_node_ids(
        prediction_tensor:torch.Tensor, 
        root:nodes.SoftmaxNode, 
        max_depth:Optional[int]=None,
        threshold:Optional[float]=None,
        progress_bar:bool=False,
    ) -> List[int]:
    """
    Takes the prediction scores for a number of samples and converts it to a list of predictions of nodes in the tree.

    Predictions use the `greedy` method which means that it chooses the greatest prediction score at each level of the tree.

    Args:
        root (SoftmaxNode): The root softmax node. Needs `set_indexes` to have been called.
        prediction_tensor (torch.Tensor): The predictions coming from the softmax layer. Shape (samples, root.layer_size)
        max_depth (int, optional): If set, then it only gives predictions at a maximum of this number of levels from the root.

    Returns:
        List[int]: A list of node IDs predicted for each sample.
    """
    prediction_nodes = greedy_predictions(
        prediction_tensor=prediction_tensor, 
        root=root, 
        max_depth=max_depth,
        threshold=threshold,
        progress_bar=progress_bar,)
    return root.get_node_ids(prediction_nodes)


def greedy_prediction_node_ids_tensor(prediction_tensor:torch.Tensor, root:nodes.SoftmaxNode, max_depth:Optional[int]=None) -> torch.Tensor:
    node_ids = greedy_prediction_node_ids(prediction_tensor=prediction_tensor, root=root, max_depth=max_depth)
    return torch.as_tensor( node_ids, dtype=int)


def render_probabilities(
        root:nodes.SoftmaxNode, 
        filepaths:List[Path]=None, 
        prediction_color="red", 
        non_prediction_color="gray",
        prediction_tensor:torch.Tensor=None,
        probabilities:torch.Tensor=None,
        predictions:List[nodes.SoftmaxNode]=None,
        horizontal:bool=True,
        threshold:float=0.005,
    ) -> List[ThresholdDotExporter]:
    """
    Renders the probabilities of each node in the tree as a graphviz graph.

    See https://anytree.readthedocs.io/en/latest/_modules/anytree/exporter/dotexporter.html for more information.

    Args:
        prediction_tensor (torch.Tensor): The output activations from the softmax layer. Shape (samples, root.layer_size)
        root (SoftmaxNode): The root softmax node. Needs `set_indexes` to have been called.
        filepaths (List[Path], optional): Paths to locations where the files can be saved. 
            Can have extension .dot or another format which can be interpreted by GraphViz such as .png or .svg. 
            Defaults to None so that files are not saved.
        prediction_color (str, optional): The color for the greedy prediction nodes and edges. Defaults to "red".
        non_prediction_color (str, optional): The color for the edges which weren't predicted. Defaults to "gray".

    Returns:
        List[DotExporter]: The list of rendered graphs.
    """
    if probabilities is None:
        assert prediction_tensor is not None, "Either `prediction_tensor` or `probabilities` must be given."
        probabilities = node_probabilities(prediction_tensor, root=root)

    if predictions is None:
        assert prediction_tensor is not None, "Either `prediction_tensor` or `node_probabilities` must be given."
        predictions = greedy_predictions(prediction_tensor, root=root)

    graphs = []
    for my_probabilities, my_prediction in zip(probabilities, predictions):
        greedy_nodes = my_prediction.ancestors + (my_prediction,)        
        graphs.append(ThresholdDotExporter(
            root, 
            probabilities=my_probabilities,
            greedy_nodes=greedy_nodes,
            horizontal=horizontal,
            prediction_color=prediction_color,
            non_prediction_color=non_prediction_color,
            threshold=threshold,
        ))

    if filepaths:
        for graph, filepath in zip(graphs, filepaths):
            filepath = Path(filepath)
            filepath.parent.mkdir(exist_ok=True, parents=True)
            if filepath.suffix == ".dot":
                graph.to_dotfile(str(filepath))
            else:
                graph.to_picture(str(filepath))

    return graphs




from sklearn.metrics import f1_score, precision_score, recall_score
import torch
from typing import Callable
from collections.abc import Sequence
from torch.nn import Module
from torch import Tensor
from torchmetrics.metric import Metric, apply_to_collection

from . import inference, nodes
from .inference import ShapeError



def target_max_depth(target_tensor:torch.Tensor, root:nodes.SoftmaxNode, max_depth:int):
    """ Converts the target tensor to the max depth of the tree. """
    if max_depth:
        max_depth_target_nodes = [root.node_list[target].path[:max_depth+1][-1] for target in target_tensor]
        target_tensor = root.get_node_ids_tensor(max_depth_target_nodes)

    return target_tensor


def greedy_accuracy(prediction_tensor, target_tensor, root, max_depth=None):
    """
    Gives the accuracy of predicting the target in a hierarchy tree.

    Predictions use the `greedy` method which means that it chooses the greatest prediction score at each level of the tree.

    Args:
        prediction_tensor (torch.Tensor): A tensor with the raw scores for each node in the tree. Shape: (samples, root.layer_size)
        target_tensor (torch.Tensor): A tensor with the target node indexes. Shape: (samples,).
        root (SoftmaxNode): The root of the hierarchy tree.

    Returns:
        float: The accuracy value (i.e. the number that are correct divided by the total number of samples)
    """    
    prediction_node_ids = inference.greedy_prediction_node_ids_tensor(prediction_tensor=prediction_tensor, root=root, max_depth=max_depth)
    target_tensor = target_max_depth(target_tensor, root, max_depth)

    return (prediction_node_ids.to(target_tensor.device) == target_tensor).float().mean()


def greedy_accuracy_depth_one(prediction_tensor, target_tensor, root):
    return greedy_accuracy(prediction_tensor, target_tensor, root, max_depth=1)
    

def greedy_accuracy_depth_two(prediction_tensor, target_tensor, root):
    return greedy_accuracy(prediction_tensor, target_tensor, root, max_depth=2)


def depth_accurate(prediction_tensor, target_tensor, root:nodes.SoftmaxNode, max_depth:int=0, threshold:float|None=None):
    """ Returns a tensor of shape (samples,) with the depth of predictions which were accurate """
    depths = []

    if root.softmax_start_index is None:
        raise nodes.IndexNotSetError(f"The index of the root node {root} has not been set. Call `set_indexes` on this object.")

    if isinstance(prediction_tensor, tuple) and len(prediction_tensor) == 1:
        prediction_tensor = prediction_tensor[0]

    if prediction_tensor.shape[-1] != root.layer_size:
        raise ShapeError(
            f"The predictions tensor given to {__name__} has final dimensions of {prediction_tensor.shape[-1]}. "
            f"That is not compatible with the root node which expects prediciton tensors to have a final dimension of {root.layer_size}."
        )

    for predictions, target in zip(prediction_tensor, target_tensor):
        node = root
        depth = 0
        target_node = root.node_list[target]
        target_path = target_node.path
        target_path_length = len(target_path)


        while (node.children):
            # This would be better if we could use torch.argmax but it doesn't work with MPS in the production version of pytorch
            # See https://github.com/pytorch/pytorch/issues/98191
            # https://github.com/pytorch/pytorch/pull/104374
            if len(node.children) == 1:
                # if this node use just one child, then we don't check the prediction
                prediction_child_index = 0
            else:
                prediction_child_index = torch.max(predictions[node.softmax_start_index:node.softmax_end_index], dim=0).indices

            node = node.children[prediction_child_index]
            depth += 1

            if depth < target_path_length and node != target_path[depth]:
                depth -= 1
                break
            
            # Stop if we have reached the maximum depth
            if max_depth and depth >= max_depth:
                break

        depths.append(depth)
    
    return torch.tensor(depths, dtype=int)


def greedy_f1_score(prediction_tensor:torch.Tensor, target_tensor:torch.Tensor, root:nodes.SoftmaxNode, average:str="macro", max_depth=None) -> float:
    """
    Gives the f1 score of predicting the target i.e. a harmonic mean of the precision and recall.

    Predictions use the `greedy` method which means that it chooses the greatest prediction score at each level of the tree.

    Args:
        prediction_tensor (torch.Tensor): A tensor with the raw scores for each node in the tree. Shape: (samples, root.layer_size)
        target_tensor (torch.Tensor): A tensor with the target node indexes. Shape: (samples,).
        root (SoftmaxNode): The root of the hierarchy tree.
        average (str, optional): The type of averaging over the different classes.
            Options are: 'micro', 'macro', 'samples', 'weighted', 'binary' or None. 
            See https://scikit-learn.org/stable/modules/generated/sklearn.metrics.f1_score.html for details.
            Defaults to "macro".

    Returns:
        float: The f1 score 
    """
    prediction_node_ids = inference.greedy_prediction_node_ids_tensor(prediction_tensor=prediction_tensor, root=root, max_depth=max_depth)
    target_tensor = target_max_depth(target_tensor, root, max_depth)

    return f1_score(target_tensor.cpu(), prediction_node_ids.cpu(), average=average)


def greedy_precision(prediction_tensor:torch.Tensor, target_tensor:torch.Tensor, root:nodes.SoftmaxNode, average:str="macro", max_depth=None) -> float:
    """
    Gives the precision score of predicting the target.

    Predictions use the `greedy` method which means that it chooses the greatest prediction score at each level of the tree.

    Args:
        prediction_tensor (torch.Tensor): A tensor with the raw scores for each node in the tree. Shape: (samples, root.layer_size)
        target_tensor (torch.Tensor): A tensor with the target node indexes. Shape: (samples,).
        root (SoftmaxNode): The root of the hierarchy tree.
        average (str, optional): The type of averaging over the different classes.
            Options are: 'micro', 'macro', 'samples', 'weighted', 'binary' or None. 
            See https://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_score.html for details.
            Defaults to "macro".

    Returns:
        float: The precision 
    """
    prediction_node_ids = inference.greedy_prediction_node_ids_tensor(prediction_tensor=prediction_tensor, root=root, max_depth=max_depth)
    target_tensor = target_max_depth(target_tensor, root, max_depth)
    
    return precision_score(target_tensor.cpu(), prediction_node_ids.cpu(), average=average)


def greedy_recall(prediction_tensor:torch.Tensor, target_tensor:torch.Tensor, root:nodes.SoftmaxNode, average:str="macro", max_depth=None) -> float:
    """
    Gives the recall score of predicting the target.

    Predictions use the `greedy` method which means that it chooses the greatest prediction score at each level of the tree.

    Args:
        prediction_tensor (torch.Tensor): A tensor with the raw scores for each node in the tree. Shape: (samples, root.layer_size)
        target_tensor (torch.Tensor): A tensor with the target node indexes. Shape: (samples,).
        root (SoftmaxNode): The root of the hierarchy tree.
        average (str, optional): The type of averaging over the different classes.
            Options are: 'micro', 'macro', 'samples', 'weighted', 'binary' or None. 
            See https://scikit-learn.org/stable/modules/generated/sklearn.metrics.recall_score.html for details.
            Defaults to "macro".

    Returns:
        float: The recall 
    """
    prediction_node_ids = inference.greedy_prediction_node_ids_tensor(prediction_tensor=prediction_tensor, root=root, max_depth=max_depth)
    target_tensor = target_max_depth(target_tensor, root, max_depth)

    return recall_score(target_tensor.cpu(), prediction_node_ids.cpu(), average=average)


def greedy_accuracy_parent(prediction_tensor, target_tensor, root, max_depth=None):
    """
    Gives the accuracy of predicting the parent of the target in a hierarchy tree.

    Predictions use the `greedy` method which means that it chooses the greatest prediction score at each level of the tree.

    Args:
        prediction_tensor (torch.Tensor): A tensor with the raw scores for each node in the tree. Shape: (samples, root.layer_size)
        target_tensor (torch.Tensor): A tensor with the target node indexes. Shape: (samples,).
        root (SoftmaxNode): The root of the hierarchy tree.

    Returns:
        float: The accuracy value (i.e. the number that are correct divided by the total number of samples)
    """    
    prediction_nodes = inference.greedy_predictions(prediction_tensor=prediction_tensor, root=root, max_depth=max_depth)
    prediction_parents = [node.parent for node in prediction_nodes]
    prediction_parent_ids = root.get_node_ids_tensor(prediction_parents)

    target_tensor = target_max_depth(target_tensor, root, max_depth)
    target_parents = [root.node_list[target].parent for target in target_tensor]
    target_parent_ids = root.get_node_ids_tensor(target_parents)

    return (prediction_parent_ids.to(target_parent_ids.device) == target_parent_ids).float().mean()


class GreedyAccuracy():
    name:str = "greedy"

    def __init__(self, root:nodes.SoftmaxNode, name="greedy_accuracy", max_depth=None):
        self.max_depth = max_depth
        self.name = name
        self.root = root

    @property
    def __name__(self):
        """ For using as a FastAI metric. """
        return self.name
    
    def __call__(self, predictions, targets):
        return greedy_accuracy(predictions, targets, self.root, max_depth=self.max_depth)


class HierarchicalSoftmaxTorchMetric(Metric):
    def _apply(self, fn: Callable, exclude_state: Sequence[str] = "") -> Module:
        """Overwrite `_apply` function such that we can also move metric states to the correct device.

        This method is called by the base ``nn.Module`` class whenever `.to`, `.cuda`, `.float`, `.half` etc. methods
        are called. Dtype conversion is guarded and will only happen through the special `set_dtype` method.

        Overriding because there is an issue device in the parent class.

        Args:
            fn: the function to apply
            exclude_state: list of state variables to exclude from applying the function, that then needs to be handled
                by the metric class itself.
        """
        this = super(Metric, self)._apply(fn)
        fs = str(fn)
        cond = any(f in fs for f in ["Module.type", "Module.half", "Module.float", "Module.double", "Module.bfloat16"])
        if not self._dtype_convert and cond:
            return this

        # Also apply fn to metric states and defaults
        for key, value in this._defaults.items():
            if key in exclude_state:
                continue

            if isinstance(value, Tensor):
                this._defaults[key] = fn(value)
            elif isinstance(value, Sequence):
                this._defaults[key] = [fn(v) for v in value]

            current_val = getattr(this, key)
            if isinstance(current_val, Tensor):
                setattr(this, key, fn(current_val))
            elif isinstance(current_val, Sequence):
                setattr(this, key, [fn(cur_v) for cur_v in current_val])
            else:
                raise TypeError(
                    f"Expected metric state to be either a Tensor or a list of Tensor, but encountered {current_val}"
                )

        # Additional apply to forward cache and computed attributes (may be nested)
        if this._computed is not None:
            this._computed = apply_to_collection(this._computed, Tensor, fn)
        if this._forward_cache is not None:
            this._forward_cache = apply_to_collection(this._forward_cache, Tensor, fn)

        return this


class GreedyAccuracyTorchMetric(HierarchicalSoftmaxTorchMetric):
    def __init__(self, root:nodes.SoftmaxNode, name:str="", max_depth=None):
        super().__init__()
        self.root = root
        self.max_depth = max_depth
        self.name = name or (f"greedy_accuracy_{max_depth}" if max_depth else "greedy_accuracy")
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")
        self.add_state("correct", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, predictions, targets):
        self.total += targets.size(0)
        self.correct += int(greedy_accuracy(predictions, targets, self.root, max_depth=self.max_depth) * targets.size(0))

    def compute(self):
        return self.correct / self.total
    

class RankAccuracyTorchMetric(HierarchicalSoftmaxTorchMetric):
    def __init__(self, root, ranks: dict[int, str], name: str = "rank_accuracy"):
        super().__init__()
        self.root = root
        self.ranks = ranks
        self.name = name

        # Use `add_state` for metrics to handle distributed reduction and device placement
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")

        for rank_name in ranks.values():
            self.add_state(rank_name, default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, predictions, targets):
        if isinstance(predictions, tuple) and len(predictions) == 1:
            predictions = predictions[0]

        # Ensure tensors match the device
        predictions = predictions.to(self.device)
        targets = targets.to(self.device)

        self.total += targets.size(0)
        depth_accurate_tensor = depth_accurate(predictions, targets, self.root)

        for depth, rank_name in self.ranks.items():
            accurate_at_depth = (depth_accurate_tensor >= depth).sum()
            setattr(self, rank_name, getattr(self, rank_name) + accurate_at_depth)

    def compute(self):
        # Compute final metric values
        return {
            rank_name: getattr(self, rank_name) / self.total
            for rank_name in self.ranks.values()
        }


class LeafAccuracyTorchMetric(HierarchicalSoftmaxTorchMetric):
    def __init__(self, root:nodes.SoftmaxNode, name:str="", max_depth=None):
        super().__init__()
        self.root = root
        self.max_depth = max_depth
        self.name = name or (f"leaf_accuracy_{max_depth}" if max_depth else "leaf_accuracy")
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")
        self.add_state("correct", default=torch.tensor(0), dist_reduce_fx="sum")
        self.node_indexes = torch.as_tensor([node.best_index_in_softmax_layer() for node in self.root.node_list])
        self.leaf_indexes = torch.as_tensor(self.root.leaf_indexes)

    def update(self, predictions, targets):
        self.total += targets.size(0)
        
        # Make sure the tensors are on the same device
        self.node_indexes = self.node_indexes.to(predictions.device)
        self.leaf_indexes = self.leaf_indexes.to(predictions.device)

        target_indices = torch.index_select(self.node_indexes.to(targets.device), 0, targets)
        
        # get indices of the maximum values along the last dimension
        probabilities = inference.leaf_probabilities(prediction_tensor=predictions, root=self.root)
        _, max_indices = torch.max(probabilities, dim=1)
        predicted_leaf_indices = torch.index_select(self.root.leaf_indexes.to(targets.device), 0, max_indices)

        self.correct += (predicted_leaf_indices == target_indices).sum()

    def compute(self):
        return self.correct / self.total
    




from torch import nn

from .nodes import SoftmaxNode
from .tensors import LazyLinearTensor

class HierarchicalSoftmaxLayerError(RuntimeError):
    pass

class HierarchicalSoftmaxLayerMixin():
    def __init__(self, root:SoftmaxNode, out_features=None, **kwargs):
        self.root = root

        if out_features is not None:
            raise HierarchicalSoftmaxLayerError(
                "Trying to create a HierarchicalSoftmaxLinearLayer by explicitly setting `out_features`. "
                "This value should be determined from the hierarchy tree and not `out_features` argument should be given to HierarchicalSoftmaxLinearLayer."
            )

        super().__init__(out_features=self.root.layer_size, **kwargs)

    def forward(self, x) -> LazyLinearTensor:
        return LazyLinearTensor(x, weight=self.weight, bias=self.bias)
    

class HierarchicalSoftmaxLinear(HierarchicalSoftmaxLayerMixin, nn.Linear):
    """
    Creates a linear layer designed to be the final layer in a neural network model that produces unnormalized scores given to HierarchicalSoftmaxLoss.

    The `out_features` value is set internally from root.layer_size and cannot be given as an argument.
    """


class HierarchicalSoftmaxLazyLinear(HierarchicalSoftmaxLayerMixin, nn.LazyLinear):
    """
    Creates a lazy linear layer designed to be the final layer in a neural network model that produces unnormalized scores given to HierarchicalSoftmaxLoss.

    The `out_features` value is set internally from root.layer_size and cannot be given as an argument.
    The `in_features` will be inferred from the previous layer at runtime.
    """



from __future__ import annotations
from anytree.exporter import DotExporter
from typing import Union
from pathlib import Path
import torch
from graphviz import Source
from anytree import Node, RenderTree, PreOrderIter, PostOrderIter, LevelOrderIter, LevelOrderGroupIter, ZigZagGroupIter
from typing import List, Optional
from rich.console import Console
console = Console()


class ReadOnlyError(RuntimeError):
    """
    Raised when trying to edit a SoftmaxNode tree after it has been set to read only.
    """


class IndexNotSetError(RuntimeError):
    """
    Raised when set_indexes not set for the SoftmaxNode root.
    """


class AlreadyIndexedError(RuntimeError):
    """
    Raised when set_indexes run more than once on a node.
    """


class SoftmaxNode(Node):
    """
    Creates a hierarchical tree to perform a softmax at each level.
    """
    def __init__(self, *args, alpha:float=1.0, weight=None, label_smoothing:float=0.0, gamma:float = None, readonly:bool=False, **kwargs):
        self.softmax_start_index = None
        self.softmax_end_index = None
        self.children_softmax_end_index = None
        self.node_to_id = None
        self.node_list = None
        self.alpha = alpha
        self.weight = weight
        self.label_smoothing = label_smoothing
        self.readonly = readonly
        self.gamma = gamma # for Focal Loss
        self.children_dict = dict()
        super().__init__(*args, **kwargs)

    def __str__(self):
        return self.name
        
    def __repr__(self):
        return str(self)
        
    def set_indexes(self, index_in_parent:Optional[int]=None, current_index:int=0) -> int:
        """
        Sets all the indexes for this node and its descendants so that each node can be referenced by the root.

        This should be called without arguments only on the root of a hierarchy tree.
        After calling this function the tree from the root down will be read only.

        Args:
            index_in_parent (int, optional): The index of this node in the parent's list of children. 
                Defaults to None which is appropriate for the root of a tree.
            current_index (int, optional): An index value for the root node to reference this node. 
                Defaults to 0 which is appropriate for the root of a tree.

        Returns:
            int: Returns the current_index
        """
        if self.softmax_start_index is not None:
            raise AlreadyIndexedError(f"Node {self} already has been indexed. It cannot be indexed again.")

        self.index_in_parent = index_in_parent
        self.index_in_parent_tensor = torch.as_tensor([index_in_parent], dtype=torch.long) if index_in_parent is not None else None
        
        self.index_in_softmax_layer = self.index_in_parent
        if self.parent:           
            # If the parent has just one child, then this node is skipped in the softmax layer because it isn't needed
            if len(self.parent.children) == 1:
                self.index_in_softmax_layer = None
            else:
                self.index_in_softmax_layer += self.parent.softmax_start_index

        if self.children:
            self.softmax_start_index = current_index
            current_index += len(self.children) if len(self.children) > 1 else 0
            self.softmax_end_index = current_index

            for child_index, child in enumerate(self.children):
                current_index = child.set_indexes(child_index, current_index)

            self.children_softmax_end_index = current_index
        
        # If this is the root, then traverse the tree and make an index of all children
        if self.parent is None:
            self.node_list = [None] * len(self.descendants)
            self.node_to_id = dict()
            non_softmax_index = self.children_softmax_end_index
            for node in self.descendants:
                if node.index_in_softmax_layer is None:
                    node_id = non_softmax_index
                    non_softmax_index += 1
                else:
                    node_id = node.index_in_softmax_layer
                
                self.node_to_id[node] = node_id
                self.node_list[node_id] = node
            
            self.node_list_softmax = self.node_list[:self.children_softmax_end_index] if self.children_softmax_end_index < len(self.node_list) else self.node_list
            self.leaf_list_softmax = [node for node in self.node_list_softmax if not node.children]
            self.node_indexes_in_softmax_layer = torch.as_tensor([node.index_in_softmax_layer for node in self.node_list_softmax])
            self.leaf_indexes = [leaf.best_index_in_softmax_layer() for leaf in self.leaves]
            try:
                self.leaf_indexes = torch.as_tensor(self.leaf_indexes, dtype=torch.long)
            except TypeError:
                pass
        
        self.readonly = True
        return current_index

    def best_index_in_softmax_layer(self) -> int|None:
        if self.index_in_softmax_layer is not None:
            return self.index_in_softmax_layer

        if self.parent:
            return self.parent.best_index_in_softmax_layer()
        
        return None

    def set_indexes_if_unset(self) -> None:
        """ 
        Calls set_indexes if it has not been called yet.
        
        This is only appropriate for the root node.
        """
        if self.root.softmax_start_index is None:
            self.root.set_indexes()

    def render(self, attr:Optional[str]=None, print:bool=False, filepath:Union[str, Path, None] = None, **kwargs) -> RenderTree:
        """
        Renders this node and all its descendants in a tree format.

        Args:
            attr (str, optional): An attribute to print for this rendering of the tree. If None, then the name of each node is used.
            print (bool): Whether or not the tree should be printed. Defaults to False.
            filepath: (str, Path, optional): A path to save the tree to using graphviz. Requires graphviz to be installed.

        Returns:
            RenderTree: The tree rendered by anytree.
        """
        rendered = RenderTree(self, **kwargs)
        if attr:
            rendered = rendered.by_attr(attr)
        if print:
            console.print(rendered)

        if filepath:
            filepath = Path(filepath)
            filepath.parent.mkdir(exist_ok=True, parents=True)

            rendered_tree_graph = DotExporter(self)
            
            if filepath.suffix == ".txt":
                filepath.write_text(str(rendered))
            elif filepath.suffix == ".dot":
                rendered_tree_graph.to_dotfile(str(filepath))
            else:
                rendered_tree_graph.to_picture(str(filepath))

        return rendered
    
    def graphviz(
        self,
        options=None,
        horizontal:bool=True,
    ) -> Source:
        """
        Renders this node and all its descendants in a tree format using graphviz.
        """
        options = options or []
        if horizontal:
            options.append('rankdir="LR";')

        dot_string = "\n".join(DotExporter(self, options=options))

        return Source(dot_string)

    def svg(
        self,
        options=None,
        horizontal:bool=True,
    ) -> str:
        """
        Renders this node and all its descendants in a tree format using graphviz.
        """
        source = self.graphviz(options=options, horizontal=horizontal)
        return source.pipe(format="svg").decode("utf-8")

    def _pre_attach(self, parent:Node):
        if self.readonly or parent.readonly:
            raise ReadOnlyError()

    def _pre_detach(self, parent:Node):
        if self.readonly or parent.readonly:
            raise ReadOnlyError()

    def _post_attach(self, parent:Node):
        """Method call after attaching to `parent`."""
        parent.children_dict[self.name] = self

    def _post_detach(self, parent:Node):
        """Method call after detaching from `parent`."""
        del parent.children_dict[self.name]

    def get_child_by_name(self, name:str) -> SoftmaxNode:
        """
        Returns the child node that has the same name as what is given.

        Args:
            name (str): The name of the child node requested.

        Returns:
            SoftmaxNode: The child node that has the same name as what is given. If not child node exists with this name then `None` is returned.
        """
        return self.children_dict.get(name, None)

    def get_node_ids(self, nodes:List) -> List[int]:
        """
        Gets the index values for descendant nodes.

        This should only be used for root nodes. 
        If `set_indexes` has been yet called on this object then it is performed as part of this function call.

        Args:
            nodes (List): A list of descendant nodes.

        Returns:
            List[int]: A list of indexes for the descendant nodes requested.
        """
        if self.node_to_id is None:
            self.set_indexes()

        return [self.node_to_id[node] for node in nodes]

    def get_node_ids_tensor(self, nodes:List) -> torch.Tensor:
        """
        Gets the index values for descendant nodes.

        This should only be used for root nodes. 
        If `set_indexes` has been yet called on this object then it is performed as part of this function call.

        Args:
            nodes (List): A list of descendant nodes.

        Returns:
            torch.Tensor: A tensor which contains the indexes for the descendant nodes requested.
        """
        return torch.as_tensor( self.get_node_ids(nodes), dtype=int)
    
    @property
    def layer_size(self) -> int:
        self.root.set_indexes_if_unset()

        return self.children_softmax_end_index
    
    def render_equal(self, string_representation:str, **kwargs) -> bool:
        """
        Checks if the string representation of this node and its descendants matches the given string.

        Args:
            string_representation (str): The string representation to compare to.
        """
        my_render = str(self.render(**kwargs))
        lines1 = str(my_render).strip().split("\n")
        lines2 = str(string_representation).strip().split("\n")

        if len(lines1) != len(lines2):
            return False

        for line1, line2 in zip(lines1, lines2):
            if line1.strip() != line2.strip():
                return False
            
        return True

    def pre_order_iter(self, depth=None, **kwargs) -> PreOrderIter:
        """ 
        Returns a pre-order iterator.
        
        See https://anytree.readthedocs.io/en/latest/api/anytree.iterators.html#anytree.iterators.preorderiter.PreOrderIter
        """
        if depth is not None:
            kwargs["maxlevel"] = depth + 1
        return PreOrderIter(self, **kwargs)
    
    def post_order_iter(self, depth=None, **kwargs) -> PostOrderIter:
        """ 
        Returns a post-order iterator.
        
        See https://anytree.readthedocs.io/en/latest/api/anytree.iterators.html#anytree.iterators.postorderiter.PostOrderIter
        """
        if depth is not None:
            kwargs["maxlevel"] = depth + 1
        return PostOrderIter(self, **kwargs)

    def level_order_iter(self, depth=None, **kwargs) -> LevelOrderIter:
        """ 
        Returns a level-order iterator.
        
        See https://anytree.readthedocs.io/en/latest/api/anytree.iterators.html#anytree.iterators.levelorderiter.LevelOrderIter
        """
        if depth is not None:
            kwargs["maxlevel"] = depth + 1
        return LevelOrderIter(self, **kwargs)

    def level_order_group_iter(self, depth=None, **kwargs) -> LevelOrderGroupIter:
        """ 
        Returns a level-order iterator with grouping starting at this node.
        
        https://anytree.readthedocs.io/en/latest/api/anytree.iterators.html#anytree.iterators.levelordergroupiter.LevelOrderGroupIter
        """
        if depth is not None:
            kwargs["maxlevel"] = depth + 1
        return LevelOrderGroupIter(self, **kwargs)

    def zig_zag_group_iter(self, depth=None, **kwargs) -> ZigZagGroupIter:
        """ 
        Returns a zig-zag iterator with grouping starting at this node.
        
        https://anytree.readthedocs.io/en/latest/api/anytree.iterators.html#anytree.iterators.zigzaggroupiter.ZigZagGroupIter
        """
        if depth is not None:
            kwargs["maxlevel"] = depth + 1
        return ZigZagGroupIter(self, **kwargs)

